#
#
import numpy as np
import os
import pickle
from netCDF4 import Dataset
import RPPtools as RPP
from mpi4py import MPI
#
import sys
sys.path.append(os.getcwd()+'/F90subroutines')
if MPI.COMM_WORLD.Get_rank() == 0 :
    err=os.system("cd F90subroutines; make all")
    if err != 0 :
        print("Compilation error in the FORTRAN interface")
        sys.exit()
else :
    print("Not compiling on other cores")
MPI.COMM_WORLD.Barrier()
#
import routing_interface
#
import configparser
config=configparser.ConfigParser({'Documentation':'false', 'nbxmax':'63'})
config.read("run.def")
gendoc=config.get("OverAll", "Documentation")
nbxmax=config.getint("OverAll", "nbxmax")
#
undef_int = 999999999.9
#
# Print the documentation for the FORTRAN interface
#
if gendoc.lower() == "true" : 
    docwrapper = open('DocumentationInterface', 'w')
    docwrapper.write(routing_interface.initatmgrid.__doc__)
    docwrapper.write("====================================================================\n")
    docwrapper.write(routing_interface.gethydrogrid.__doc__)
    docwrapper.write("====================================================================\n")
    docwrapper.write(routing_interface.findbasins.__doc__)
    docwrapper.write("====================================================================\n")
    docwrapper.write(routing_interface.globalize.__doc__)
    docwrapper.write("====================================================================\n")
    docwrapper.write(routing_interface.linkup.__doc__)
    docwrapper.write("====================================================================\n")
    docwrapper.write(routing_interface.fetch.__doc__)
    docwrapper.write("====================================================================\n")
    docwrapper.write(routing_interface.truncate.__doc__)
    docwrapper.close
#
# Functions to access the interfaces
#
#
# initatmgrid : Initialises the grid.f90 module and passes the description of the atmospheric grid.
#
def initatmgrid(nbpt, modelgrid) :
    print("INITATMGRID corners", np.array(modelgrid.polyll).shape)
    print("INITATMGRID area", np.array(modelgrid.area).shape)
    print("INITATMGRID neighbours", np.array(modelgrid.neighbours).shape)
    routing_interface.initatmgrid( modelgrid.polyll, modelgrid.coordll, modelgrid.area, modelgrid.contfrac, modelgrid.neighbours)
    return
#
#
class HydroOverlap :
#
    def __init__(self, nbpt, nbvmax, sub_pts, sub_index_in, sub_area_in, sub_lon_in, sub_lat_in,  modelgrid, hydrodata) :
        #
        # Reshape stuff so that it fits into arrays
        #
        sub_index = np.zeros((nbpt,nbvmax,2), dtype=np.int8, order='F')
        sub_area = np.zeros((nbpt,nbvmax), dtype=np.float32, order='F')
        sub_lon = np.zeros((nbpt,nbvmax), dtype=np.float32, order='F')
        sub_lat = np.zeros((nbpt,nbvmax), dtype=np.float32, order='F')
        for ib in range(nbpt) :
            sub_area[ib,0:sub_pts[ib]] = sub_area_in[ib][:]
            sub_lon[ib,0:sub_pts[ib]] = sub_lon_in[ib][:]
            sub_lat[ib,0:sub_pts[ib]] = sub_lat_in[ib][:]
            for ip in range(sub_pts[ib]) :
                sub_index[ib,ip,:] = [sub_index_in[ib][0][ip],sub_index_in[ib][1][ip]]
        #
        trip_tmp = np.zeros((nbpt,nbvmax), dtype=np.float32, order='F')
        basins_tmp = np.zeros((nbpt,nbvmax), dtype=np.float32, order='F')
        topoind_tmp = np.zeros((nbpt,nbvmax), dtype=np.float32, order='F')
        fac_tmp = np.zeros((nbpt,nbvmax), dtype=np.float32, order='F')
        hierarchy_tmp = np.zeros((nbpt,nbvmax), dtype=np.float32, order='F')
        #
        trip_tmp[:,:] = np.nan
        basins_tmp[:,:] = np.nan
        for ib in range(nbpt) :
            trip_tmp[ib,0:sub_pts[ib]] = np.asarray(hydrodata.trip[ib][:])
            basins_tmp[ib,0:sub_pts[ib]] = np.asarray(hydrodata.basins[ib][:])
            topoind_tmp[ib,0:sub_pts[ib]] = np.asarray(hydrodata.topoind[ib][:])
            fac_tmp[ib,0:sub_pts[ib]] = np.asarray(hydrodata.fac[ib][:])
            hierarchy_tmp[ib,0:sub_pts[ib]] = np.asarray(hydrodata.disto[ib][:])
        #
        trip_tmp[np.isnan(trip_tmp)] = undef_int
        basins_tmp[np.isnan(trip_tmp)] = undef_int
        #
        # Go to the call of the FORTRAN interface
        #
        print("GETHYDROGRID : nbpt = ", nbpt, nbvmax)
        print("GETHYDROGRID : nbvmax = ", nbvmax)
        print("GETHYDROGRID : nbxmax = ", nbxmax)
        self.nbi, self.nbj, self.area_bx, self.trip_bx, self.basin_bx, self.topoind_bx, self.fac_bx, self.hierarchy_bx, \
            self.lon_bx, self.lat_bx, self.lshead_bx, self.nwbas = \
                    routing_interface.gethydrogrid(nbxmax, sub_pts, sub_index, sub_area, \
                    hydrodata.basinsmax, hydrodata.topoindmin, sub_lon, sub_lat, trip_tmp, basins_tmp, topoind_tmp, fac_tmp, hierarchy_tmp)
        #
        # Plot some diagnostics for the hydrology grid within the atmospheric meshes.
        #
        # Clean-up these arrays so that they are easy to use in Python.
        self.lon_bx[self.lon_bx > 360.]=np.nan
        self.lat_bx[self.lat_bx > 90.]=np.nan
        # with open("Grids.txt", 'wb') as fp:
        #     pickle.dump(lon_bx, fp)
        #     pickle.dump(lat_bx, fp)
        #
        
class HydroSuper :
    def __init__(self, nbvmax, hydrodata, hydrooverlap) :
        #
        # Call findbasins
        #
        nb_basin, basin_inbxid, basin_outlet, basin_outtp, self.basin_sz, basin_bxout, basin_bbout, self.basin_pts, basin_lshead, coast_pts = \
                    routing_interface.findbasins(nbvmax, hydrooverlap.nbi, hydrooverlap.nbj, hydrooverlap.trip_bx, \
                                                 hydrooverlap.basin_bx, hydrooverlap.fac_bx, hydrooverlap.hierarchy_bx, \
                                                 hydrooverlap.topoind_bx, hydrooverlap.lshead_bx, \
                                                 hydrooverlap.lon_bx, hydrooverlap.lat_bx)
        #
        # Call Globalize
        #
        self.basin_count, self.basin_notrun, self.basin_area, self.basin_hierarchy, self.basin_fac, self.basin_topoind, \
            self.basin_id, self.basin_coor, self.basin_type, self.basin_flowdir, \
            self.basin_lshead, self.outflow_grid, self.outflow_basin, self.nbcoastal, self.coastal_basin = \
                    routing_interface.globalize(hydrooverlap.area_bx, hydrooverlap.trip_bx, hydrooverlap.hierarchy_bx,\
                                                hydrooverlap.fac_bx, hydrooverlap.topoind_bx, hydrodata.topoindmin, nb_basin, \
                                                basin_inbxid, basin_outlet, basin_outtp, self.basin_sz, self.basin_pts, basin_bxout, \
                                                basin_bbout, basin_lshead, coast_pts, hydrooverlap.nwbas)
        #
    def linkup(self, hydrodata) :
        #
        # Call the linkup routine in routing_reg.
        #
        print("Invented basins =", hydrodata.basinsmax)
        self.inflow_number,self.inflow_grid,self.inflow_basin = routing_interface.linkup(nbxmax, self.basin_count, self.basin_area, self.basin_id, \
                                                                       self.basin_flowdir, self.basin_lshead, self.basin_hierarchy, \
                                                                       self.basin_fac, self.outflow_grid, self.outflow_basin, \
                                                                       self.nbcoastal, self.coastal_basin, float(hydrodata.basinsmax))
    def fetch(self) :
        self.fetch_basin = routing_interface.fetch(self.basin_count, self.basin_area, self.basin_id, self.basin_hierarchy, \
                                                   self.basin_fac, self.outflow_grid, self.outflow_basin)


class HydroGraph :
    def __init__(self, nbasmax, hydrosuper) :
        self.nbasmax = nbasmax
        self.routing_area, self.topo_resid, self.route_togrid, self.route_tobasin, self.route_nbintobas, self.global_basinid, \
            self.route_outlet, self.route_type, self.origin_nbintobas = \
                                    routing_interface.truncate(nbasmax, hydrosuper.basin_count, hydrosuper.basin_notrun, hydrosuper.basin_area, \
                                                                hydrosuper.basin_topoind, hydrosuper.fetch_basin, hydrosuper.basin_id, \
                                                                hydrosuper.basin_coor, hydrosuper.basin_type, hydrosuper.basin_flowdir, \
                                                                hydrosuper.outflow_grid, hydrosuper.outflow_basin, \
                                                                hydrosuper.inflow_number,hydrosuper.inflow_grid,hydrosuper.inflow_basin)

    def dumpnetcdf(self, filename, globalgrid, procgrid, part) :
        #
        NCFillValue=1.0e20
        vtyp=np.float64
        cornerind=[0,2,4,6]
        #
        if part.rank == 0 :
            outnf=Dataset(filename, 'w', format='NETCDF4_CLASSIC')
            # Dimensions
            outnf.createDimension('x', globalgrid.ni)
            outnf.createDimension('y', globalgrid.nj)
            outnf.createDimension('land', len(procgrid.area))
            outnf.createDimension('htu', self.nbasmax)
            outnf.createDimension('bnd', 4)
        #
        # Coordinates
        #
        # Longitude
        longitude = part.gather(procgrid.lon_full)
        if part.rank == 0 :
            lon=outnf.createVariable("lon", vtyp, ('y','x'), fill_value=NCFillValue)
            lon.units="grid box centre degrees east"
            lon.title="Longitude"
            lon.axis="X"
            lon[:,:] = longitude[:,:]
        # # Longitude bounds
        # lonbnd=outnf.createVariable("lon_bnd", vtyp, ('bnd','y','x'), fill_value=NCFillValue)
        # lonbnd.units="grid box corners degrees east"
        # lonbnd.title="Longitude Corners"
        # lonbnd[:,:] = np.array(procgrid.polyll)[:,cornerind,0]
        #
        # Latitude
        latitude = part.gather(procgrid.lat_full)
        if part.rank == 0 :
            lat=outnf.createVariable("lat", vtyp, ('y','x'), fill_value=NCFillValue)
            lat.units="grid box centre degrees north"
            lat.standard_name="grid latitude"
            lat.title="Latitude"
            lat.axis="Y"
            lat[:] = latitude[:,:]
        # # Latitude bounds
        # latbnd=outnf.createVariable("lat_bnd", vtyp, ('bnd','land'), fill_value=NCFillValue)
        # latbnd.units="grid box corners degrees north"
        # latbnd.title="Latitude Corners"
        # latbnd[:,:] = np.array(procgrid.polyll)[:,cornerind,1]
        #
        #
        # Land sea mask
        #
        if part.rank == 0 :
                land=outnf.createVariable("land", vtyp, ('y','x'), fill_value=NCFillValue)
                land.units="Land Sea mask"
                land.standard_name="landsea mask"
                land.title="Land"
                land[:,:] = globalgrid.land[:,:]
        # Area
        areas = procgrid.landscatter(np.array(procgrid.area, dtype=np.float64))
        areas[np.isnan(areas)] = NCFillValue
        if part.rank == 0 :
            area=outnf.createVariable("area", vtyp, ('y','x'), fill_value=NCFillValue)
            area.units="m^2"
            area.standard_name="grid area"
            area.title="Area"
        else :
            area = np.zeros((1,1))
        area[:,:] = part.gather(areas[:,:])
        #
        # Variables
        # Once gathered on root_proc we transform them into float64. Careful, Integer variables do not have NaN ! 
        #
        rarea = procgrid.landscatter(self.routing_area[:,:], order='F')
        rarea = rarea.astype(np.float64, copy=False)
        rarea[np.isnan(rarea)] = NCFillValue
        if part.rank == 0 :
            routingarea = outnf.createVariable("routingarea", vtyp, ('htu','y','x'), fill_value=NCFillValue)
            routingarea.title = "Surface of basin"
            routingarea.units = "m^2"
        else :
            routingarea = np.zeros((1,1,1))
        routingarea[:,:,:] = part.gather(rarea)
        #
        rgrid = procgrid.landscatter(self.route_togrid[:,:], order='F')
        rgrid = rgrid.astype(np.float64, copy=False)
        rgrid[rgrid >= RPP.IntFillValue] = NCFillValue
        if part.rank == 0 :
            routetogrid = outnf.createVariable("routetogrid", vtyp, ('htu','y','x'), fill_value=NCFillValue)
            routetogrid.title = "Grid into which the basin flows"
            routetogrid.units = "-"
        else :
            routetogrid = np.zeros((1,1,1))    
        routetogrid[:,:,:] = part.gather(rgrid)
        #
        rtobasin = procgrid.landscatter(self.route_tobasin[:,:], order='F')
        rtobasin = rtobasin.astype(np.float64, copy=False)
        rtobasin[rtobasin >= RPP.IntFillValue] = NCFillValue
        if part.rank == 0 :
            routetobasin = outnf.createVariable("routetobasin", vtyp, ('htu','y','x'), fill_value=NCFillValue)
            routetobasin.title = "Basin in to which the water goes"
            routetobasin.units = "-"
        else :
            routetobasin = np.zeros((1,1,1))
        routetobasin[:,:,:] = part.gather(rtobasin)
        #
        rid = procgrid.landscatter(self.global_basinid[:,:], order='F')
        rid = rid.astype(np.float64, copy=False)
        rid[rid >= RPP.IntFillValue] = NCFillValue
        if part.rank == 0 :                           
            basinid = outnf.createVariable("basinid", vtyp, ('htu','y','x'), fill_value=NCFillValue)
            basinid.title = "ID of basin"
            basinid.units = "-"
        else :
            basinid = np.zeros((1,1,1))
        basinid[:,:,:] = part.gather(rid)
        #
        rintobas = procgrid.landscatter(self.route_nbintobas[:])
        rintobas = rintobas.astype(np.float64, copy=False)
        rintobas[rintobas >= RPP.IntFillValue] = NCFillValue
        if part.rank == 0 : 
            routenbintobas = outnf.createVariable("routenbintobas", vtyp, ('y','x'), fill_value=NCFillValue)
            routenbintobas.title = "Number of basin into current one"
            routenbintobas.units = "-"
        else :
            routenbintobas = np.zeros((1,1))
        routenbintobas[:,:] = part.gather(rintobas)
        #
        onbintobas = procgrid.landscatter(self.origin_nbintobas[:])
        onbintobas = onbintobas.astype(np.float64, copy=False)
        onbintobas[onbintobas >= RPP.IntFillValue] = NCFillValue
        if part.rank == 0 :
            originnbintobas = outnf.createVariable("originnbintobas", vtyp, ('y','x'), fill_value=NCFillValue)
            originnbintobas.title = "Number of sub-grid basin into current one before truncation"
            originnbintobas.units = "-"
        else :
            originnbintobas = np.zeros((1,1))
        originnbintobas[:,:] = part.gather(onbintobas)
        #
        olat = procgrid.landscatter(self.route_outlet[:,:,0], order='F')
        olat = olat.astype(np.float64, copy=False)
        olat[np.isnan(olat)] = NCFillValue
        if part.rank == 0 :
            outletlat = outnf.createVariable("outletlat", vtyp, ('htu','y','x'), fill_value=NCFillValue)
            outletlat.title = "Latitude of Outlet"
            outletlat.title = "degrees north"
        else :
            outletlat = np.zeros((1,1,1))
        outletlat[:,:,:] = part.gather(olat)
        #
        olon = procgrid.landscatter(self.route_outlet[:,:,1], order='F')
        olon = olon.astype(np.float64, copy=False)
        olon[np.isnan(olon)] = NCFillValue
        if part.rank == 0 :
            outletlon = outnf.createVariable("outletlon", vtyp, ('htu','y','x'), fill_value=NCFillValue)
            outletlon.title = "Longitude of outlet"
            outletlon.units = "degrees east"
        else :
            outletlon = np.zeros((1,1,1))
        outletlon[:,:,:] = part.gather(olon)
        #
        otype = procgrid.landscatter(self.route_type[:,:], order='F')
        otype = otype.astype(np.float64, copy=False)
        otype[np.isnan(otype)] = NCFillValue
        if part.rank == 0 :
            outlettype = outnf.createVariable("outlettype", vtyp, ('htu','y','x'), fill_value=NCFillValue)
            outlettype.title = "Type of outlet"
            outlettype.units = "code"
        else :
            outlettype = np.zeros((1,1,1))
        outlettype[:,:,:] = part.gather(otype)
        #
        tind = procgrid.landscatter(self.topo_resid[:,:], order='F')
        tind = tind.astype(np.float64, copy=False)
        tind[np.isnan(tind)] = NCFillValue
        if part.rank == 0 :
            topoindex = outnf.createVariable("topoindex", vtyp, ('htu','y','x'), fill_value=NCFillValue)
            topoindex.title = "Topographic index of the retention time"
            topoindex.units = "m"
        else :
            topoindex = np.zeros((1,1,1))
        topoindex[:,:,:] = part.gather(tind)
        #
        if part.rank == 0 :
            outnf.close()





